//
// Member functions for the compact data type implement a static Huffman coder.
// John Robinson.
//
#include <string.h>
#include "compact.h"
// #define DEBUG 1
// #define DEBUG2 1

void compactor::reset(void) {
#ifdef DEBUG
	if (debug)
		printf("\nCoder %d: reset:\n",id); fflush(stdout);
#endif
	for (int i = 0; i < 256; i++)
		freq[i] = 0;
	}

void compactor::buildtree(void)
// Based on JPEG huffman algorithm
	{
#ifdef DEBUG
	if (debug)
		printf("\nCoder %d: buildtree:\n",id); fflush(stdout);
#endif
	int i;
	int h1;
	int h2;
	int length[256];
	int descendents[256];
/*
** Debugging
*/
	accum = 0;
	maxval = 0;
	for (i = 0; i < 256; i++) {
		savefreq[i] = freq[i];
		accum += freq[i];
		if (freq[i] > 0)
			maxval = i;
		}
	if (accum < 1)
		accum = 1;

	
	for (i = 0; i < 256; i++)
		{
		length[i] = 0;
		descendents[i] = -1;
		}
	while(1) {
		h1 = -1;
		h2 = -1;
		// Find smallest non-zero frequency
		for (i = 0; i < 256; i++) {
			if (freq[i]) {
			    if (h1 == -1 || freq[i] < freq[h1]) {
				if (h2 == -1 || freq[h1] < freq[h2])
					h2 = h1;
				h1 = i;
			    } else if (h2 == -1 || freq[i] < freq[h2])
				h2 = i;
			}
		}
		if (h2 == -1) // Found the root
			break;
		// Merge the two trees
		freq[h1] += freq[h2];
		freq[h2] = 0;
		length[h1]++;
		while (descendents[h1] >= 0) {
			h1 = descendents[h1];	// Go down chain
			length[h1]++;
			}
		//  Put h2 on this chain
		descendents[h1] = h2;
		// And go down its chain
		while (descendents[h1] >= 0) {
			h1 = descendents[h1];
			length[h1]++;
			}
		}
	if (h1 == -1) // Empty tree - create dummy char
		{
		longest = 0;
		num_symbols_of_length[0] = 1;
		ordered_symbols[0] = 0;
		return;
		}
	if (length[h1] == 0)	// one symbol
		{
		longest = 0;
		num_symbols_of_length[0] = 1;
		ordered_symbols[0] = h1;
		return;
		}
	for (i = 0; i < 32; i++)
		num_symbols_of_length[i] = 0;
	for (i = 0; i < 256; i++) {
		if (length[i])
			num_symbols_of_length[length[i]]++;
			// Assume never have case of 256 equal symbols
		}
	// Reduce to TABLEBITS-sized words using JPEG algorithm
	for (i = 31; i > TABLEBITS; i--) {
		while(num_symbols_of_length[i]) {
			int j = i - 2;
			while (!num_symbols_of_length[j])
				j--;
			num_symbols_of_length[i] -= 2;
			num_symbols_of_length[i-1]++;
			num_symbols_of_length[j+1] += 2;
			num_symbols_of_length[j]--;
			}
		}
	while(num_symbols_of_length[i] == 0)
		i--;
	// i is now length of longest codeword
	longest = i;
	int k = 0;
	for (i = 1; i <= 31; i++) {
		for (int j = 0; j < 256; j++)
			if (length[j] == i)
				ordered_symbols[k++] = j;
		}
	}

// Decide whether to use Huffman coding or not.
// This function is coded as a mirror of writetree, with the stuff that's
// done in debugging in writetree elevated to "real" functionality, and
// the writing to file commented out
int compactor::usetree(void) {
#ifdef DEBUG
	if (debug) {
		printf("\nCoder %d: usetree:\n",id); fflush(stdout);
		printf("Total symbols to code = %d\n", accum);
		}
#endif
	int litbits;
	int codelength = 0;
	int tablelength = 0;
	int i;
	// pfio->putbit(longest & 0x8);
	// pfio->putbit(longest & 0x4);
	// pfio->putbit(longest & 0x2);
	// pfio->putbit(longest & 0x1);
	tablelength += 4;
#ifdef DEBUG
	if (debug) {
		printf("longest = %d\n", longest); fflush(stdout);
		}
#endif
	// Work out the longest literal
	if (maxval > 127) litbits = 8;
	else if (maxval > 63) litbits = 7;
	else if (maxval > 31) litbits = 6;
	else if (maxval > 15) litbits = 5;
	else if (maxval > 7) litbits = 4;
	else if (maxval > 3) litbits = 3;
	else if (maxval > 1) litbits = 2;
	else if (maxval > 0) litbits = 1;
	else litbits = 0;
	litlength = litbits;
	// pfio->putbit(litbits & 0x8);
	// pfio->putbit(litbits & 0x4);
	// pfio->putbit(litbits & 0x2);
	// pfio->putbit(litbits & 0x1);
	// tablelength += 4;
	// Don't add onto table length here because we have this cost
	// if we decide to use literals too.
	for (i = 1; i <= longest; i++)
		{
#ifdef DEBUG
		if (debug) {
			printf("num of length %d = %d\n",i, num_symbols_of_length[i]);
			}
#endif
		// pfio->putbyte(num_symbols_of_length[i]);
		tablelength += 8;
		}
	// Then the literals
	int k = 0;
	for (i = 0; i <= longest; i++) {
#ifdef DEBUG
		if (debug) {
			printf("%d bits: ", i);
			}
#endif
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			{
#ifdef DEBUG
			if (debug) {
				printf("%d[n=%d] ", ordered_symbols[k],
				savefreq[ordered_symbols[k]]);
				}
#endif
			codelength += i*savefreq[ordered_symbols[k]];
			if (litbits) {
				// pfio->putbits(ordered_symbols[k++],litbits);
				k++;
				tablelength += litbits;
				}
			}
#ifdef DEBUG
		if (debug)
			printf("\n");
#endif
		}
#ifdef DEBUG
    if (debug) {
	printf("Huffman code will code %d symbols in %d bits (%d bytes).\n",
		accum, codelength, (codelength+7)/8);
	printf("To transmit code table required %d bits (%d bytes).\n",
		tablelength, (tablelength+7)/8);
	}
#endif
	codelength += tablelength;
#ifdef DEBUG
    if (debug) {
	printf("Total bits required (table+code) = %d (%d bytes).\n",
		codelength, (codelength+7)/8);
	printf("Average rate = %f bits/symbol\n",(float)codelength/accum);
	printf("Maximum input symbol value is %d.\n", maxval);
	printf("Sending %d-bit literals directly requires %d bits (%d bytes).\n",
		litbits, litbits*accum, (litbits*accum+7)/8);
	}
#endif
	litbits *= accum;
#ifdef DEBUG
    if (debug) {
	if (litbits > codelength)
		printf("++Huffman coding gives saving of %d bits (%d bytes).\n",
			litbits-codelength, (litbits-codelength+7)/8);
	else
		printf("--Huffman coding adds additional %d bits (%d bytes).\n",
			codelength-litbits, (codelength-litbits+7)/8);
	printf("\n");
        }
#endif
	if (litbits < codelength)
		return 0;	// Don't use Huffman
	return 1;
	}

// Write the Huffman tree to the output code file
// Most of the debug stuff is now commented out because it is done in usetree.
// There is some superfluous stuff left behind though in case we ever want
// to go back to not having usetree.
void compactor::writetree(void) {
#ifdef DEBUG
	if (debug) {
		printf("\nCoder %d: writetree:\n",id); fflush(stdout);
		printf("Total symbols to code = %d\n", accum);
		}
#endif
	if (usetree())
		usinghuffman = 1;
	else
		usinghuffman = 0;
#ifdef DEBUG
	if (debug) {
		if (usinghuffman)
			printf("Using Huffman:\n");
		else
			printf("Not Using Huffman:\n");
		}
#endif
	pfio->putbit(usinghuffman);
	int litbits;
	int codelength = 0;
	int tablelength = 0;
	int i;
	if (usinghuffman) {
		pfio->putbit(longest & 0x8);
		pfio->putbit(longest & 0x4);
		pfio->putbit(longest & 0x2);
		pfio->putbit(longest & 0x1);
/*
#ifdef DEBUG
		if (debug) {
			printf("longest = %d\n", longest); fflush(stdout);
			}
#endif
*/
		}
	tablelength += 4;
	// Work out the longest literal
	if (maxval > 127) litbits = 8;
	else if (maxval > 63) litbits = 7;
	else if (maxval > 31) litbits = 6;
	else if (maxval > 15) litbits = 5;
	else if (maxval > 7) litbits = 4;
	else if (maxval > 3) litbits = 3;
	else if (maxval > 1) litbits = 2;
	else if (maxval > 0) litbits = 1;
	else litbits = 0;
	pfio->putbit(litbits & 0x8);
	pfio->putbit(litbits & 0x4);
	pfio->putbit(litbits & 0x2);
	pfio->putbit(litbits & 0x1);
/*
#ifdef DEBUG
	if (debug) {
		printf("litbits = %d\n", litbits); fflush(stdout);
		}
#endif
*/
	if (!usinghuffman)
		return;
	tablelength += 4;
	for (i = 1; i <= longest; i++)
		{
/*
#ifdef DEBUG
		if (debug) {
			printf("num of length %d = %d\n",i, num_symbols_of_length[i]);
			}
#endif
*/
		pfio->putbyte(num_symbols_of_length[i]);
		tablelength += 8;
		}
	// Then the literals
	int k = 0;
	for (i = 0; i <= longest; i++) {
/*
#ifdef DEBUG
		if (debug) {
			printf("%d bits: ", i);
			}
#endif
*/
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			{
/*
#ifdef DEBUG
			if (debug) {
				printf("%d[n=%d] ", ordered_symbols[k],
				savefreq[ordered_symbols[k]]);
				}
#endif
*/
			codelength += i*savefreq[ordered_symbols[k]];
			if (litbits) {
				pfio->putbits(ordered_symbols[k++],litbits);
				tablelength += litbits;
				}
			}
/*
#ifdef DEBUG
		if (debug)
			printf("\n");
#endif
*/
		}
	fflush(stdout);
/*
#ifdef DEBUG
    if (debug) {
	printf("Huffman code will code %d symbols in %d bits (%d bytes).\n",
		accum, codelength, (codelength+7)/8);
	printf("To transmit code table required %d bits (%d bytes).\n",
		tablelength, (tablelength+7)/8);
	codelength += tablelength;
	printf("Total bits required (table+code) = %d (%d bytes).\n",
		codelength, (codelength+7)/8);
	printf("Average rate = %f bits/symbol\n",(float)codelength/accum);
	printf("Maximum input symbol value is %d.\n", maxval);
	printf("Sending %d-bit literals directly requires %d bits (%d bytes).\n",
		litbits, litbits*accum, (litbits*accum+7)/8);
	litbits *= accum;
	if (litbits > codelength)
		printf("++Huffman coding gives saving of %d bits (%d bytes).\n",
			litbits-codelength, (litbits-codelength+7)/8);
	else
		printf("--Huffman coding adds additional %d bits (%d bytes).\n",
			codelength-litbits, (codelength-litbits+7)/8);
	printf("\n");
    }
#endif
*/
	maketable_encode();
	}

void compactor::copytree(compactor& source) {
#ifdef DEBUG
	if (debug) {
		printf("\nCoder %d: copytree:\n",id);
		}
#endif
	longest = source.longest;
	int i, j, k;
	for (i = 0; i <= longest; i++) 
		num_symbols_of_length[i] = source.num_symbols_of_length[i];
	k = 0;
	// printf("Transferring symbols\n");
	for (i = 0; i <= longest; i++) {
		// printf("Length %d:\n", i);
		for (j = 0; j < num_symbols_of_length[i]; j++) {
			ordered_symbols[k] = source.ordered_symbols[k];
			k++;
			// printf("%d ", ordered_symbols[k-1]);
			}
		printf("\n");
		}
	}
int compactor::copytreewithout(compactor& source, int from, int to) {
#ifdef DEBUG
	if (debug) {
		printf("\nCoder %d: copytreewithout(%d,%d):\n",id,from,to);
		}
#endif
	int totalsymbols;
	longest = source.longest;
	int i, j, k;
	for (i = 0; i <= longest; i++) 
		num_symbols_of_length[i] = source.num_symbols_of_length[i];
	k = 0;
	// printf("Transferring symbols\n");
	for (i = 0; i <= longest; i++) {
		// printf("Length %d:\n", i);
		for (j = 0; j < num_symbols_of_length[i]; j++) {
			ordered_symbols[k] = source.ordered_symbols[k];
			k++;
			// printf("%d ", ordered_symbols[k-1]);
			}
		// printf("\n");
		}
	totalsymbols = k;
	// printf("Removing symbols\n");
	while(1) {  // Will leave when no more symbols to check
		int temp = from - 1;
		k = 0;	// Restart at beginning each time until we do
			// cleverer algorithm to speed up.
		for (i = 0; i <= longest; i++) {
			for (j = 0; j < num_symbols_of_length[i]; j++) {
				temp = ordered_symbols[k];
				if ((temp >= from)&&(temp <= to))
					break;
				k++;
				}
			if ((temp >= from)&&(temp <= to))
				break;
			}
		if (i > longest)
			break; // Leave while(1)
		// First remove the symbol
		// printf("Found symbol %d of length %d at position %d\n", 
		// 				temp, i, k);
		int l = k+1;
		while (l < totalsymbols) {
			ordered_symbols[l-1] = ordered_symbols[l];
			l++;
			}
		totalsymbols--;
		num_symbols_of_length[i]--;
		int numtopromote = 1;
		while ((i <= longest)&&numtopromote) {
			while((num_symbols_of_length[i])&&(numtopromote)) {
				num_symbols_of_length[i-1]++;
				num_symbols_of_length[i]--;
				// printf("Promoting a symbol of length %d\n", i);
				numtopromote--;
				}
			// Promoted as many as we can of this length
			numtopromote *= 2; // Because going down level
			i++;
			}
		if (num_symbols_of_length[longest] == 0) {
			// printf("Length of longest code reduced\n");
			longest--;
			}
		}
/*
	k = 0;
	printf("Revised lengths\n");
	for (i = 0; i <= longest; i++) {
		printf("Length %d:\n", i);
		for (int j = 0; j < num_symbols_of_length[i]; j++) {
			printf("%d ", ordered_symbols[k++]);
			}
		printf("\n");
		}
*/
	return 0;
	}
// Read the Huffman tree from the input code file.
void compactor::readtree(void) {
	int i;
	usinghuffman = pfio->getbit();
	if (usinghuffman) {
		longest = pfio->lookbits(4,0xf);
		pfio->skipbits(4);
		}
	int litbits = pfio->lookbits(4,0xf);
	int litmask = (1<<litbits) - 1;
	pfio->skipbits(4);
	litlength = litbits;
#ifdef DEBUG
    if (debug) {
	printf("\nCoder %d: readtree:\n",id);
	if (usinghuffman)
		printf("Huffman: longest codeword = %d\n", longest);
	else
		printf("Not using Huffman\n");
	printf("litbits = %d\n", litbits); fflush(stdout);
	}
#endif
	if (!usinghuffman)
		return;
	if (!longest) {
		num_symbols_of_length[0] = 1;
		ordered_symbols[0] = 0;
		}
	else
		num_symbols_of_length[0] = 0;
	for (i = 1; i <= longest; i++) {
		num_symbols_of_length[i] = pfio->lookbits(8,0xff);
		pfio->skipbits(8);
#ifdef DEBUG
		if (debug)
			printf("num symbols of length %d = %d\n",i,num_symbols_of_length[i]);
#endif
		}
	int k = 0;
	for (i = 0; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++) {
			ordered_symbols[k++] = pfio->lookbits(litbits,litmask);
			pfio->skipbits(litbits);
#ifdef DEBUG
			if (debug)
				printf("%d ", ordered_symbols[k-1]);
#endif
			}
		}
#ifdef DEBUG
	if (debug) {
		printf("\n"); fflush(stdout);
		}
#endif
	maketable_decode();
	}
void compactor::maketable_encode(void) {
	// Create coding tables
	// First create temporary tables ordered by code length
#ifdef DEBUG
	// printf("(maketable_encode called by writetree:)\n", id);
#endif
	unsigned char codelength[257];
	unsigned short codebits[256];
	int k = 0;
	for (int i = 1; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			codelength[k++] = i;
		}
	codelength[k] = 0;	// Sentinal
	int lastsymbol = k;
	int codeword = 0;
	k = 0;
	int total = codelength[0];
	while(codelength[k]) {
		while(codelength[k] == total) {
			codebits[k++] = codeword;
			codeword++;
			}
		codeword <<= 1;
		total++;
		}
	// And finally install in lookup table
	memset(enclength,0,256);	// So all unused slots will be 0
	if (!longest) {
#ifdef DEBUG
		// printf("0-bit symbol: %d\n",ordered_symbols[0]); fflush(stdout);
#endif
		}
	for (k = 0; k < lastsymbol; k++) {
		encode[ordered_symbols[k]] = codebits[k];
		enclength[ordered_symbols[k]] = codelength[k];
#ifdef DEBUG
/*
		printf("%d: ",ordered_symbols[k]);
		for (int temp = 1; temp <= codelength[k]; temp++) {
			if (codebits[k] & (1<<(codelength[k]-temp)))
				printf("1");
			else
				printf("0");
			}
		printf("\n"); fflush(stdout);
*/
#endif
		}
	}
void compactor::maketable_decode(void) {
	// Create coding tables
#ifdef DEBUG
	if (debug) {
		printf("(maketable_decode called by readtree):\n"); fflush(stdout);
		}
#endif
	int tablesize = 1<<longest;
	longmask = tablesize - 1;
	if (!longest)	// Null table
		{
		memset(declength,0,TABLESIZE);
		memset(decode,ordered_symbols[0],TABLESIZE);
#ifdef DEBUG
		if (debug) {
			printf("Null table with single 0-bit symbol: %d\n",ordered_symbols[0]); fflush(stdout);
			}
#endif
		return;
		}
	// First create temporary tables ordered by code length
	unsigned char codelength[257];
	unsigned short codebits[256];
	int k = 0;
	int i;
	for (i = 1; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			codelength[k++] = i;
		}
	codelength[k] = 0;	// Sentinal
	int codeword = 0;
	k = 0;
	int total = codelength[0];
	while(codelength[k]) {
		while(codelength[k] == total) {
			codebits[k++] = codeword;
			codeword++;
			}
		codeword <<= 1;
		total++;
		}
	// And finally install in lookup table
	k = 0;
	for (int length = 1; length <= longest; length++) {
		for (i = 1; i <= num_symbols_of_length[length]; i++) {
			int entry = codebits[k] << (longest-length);
			int j = 1 << (longest-length);
#ifdef DEBUG
		if (debug) {
		    printf("%d: %d bits: ",ordered_symbols[k],codelength[k]);
		    printf("installing in table from %d to %d... ",
			entry, entry+j);
		    for (int temp = 1; temp <= codelength[k]; temp++) {
			if (codebits[k] & (1<<(codelength[k]-temp)))
				printf("1");
			else
				printf("0");
			}
		    printf("\n"); fflush(stdout);
		    }
#endif
			while(j--) {
				decode[entry] = ordered_symbols[k];
				declength[entry] = length;
				entry++;
				}
			k++;
			}
		}
	}
void compactor::put(int symbol)
{
	int cnt;
	if (!usinghuffman) {
		pfio->putbits(symbol,litlength);
#ifdef DEBUG
	if (debug)
		printf("Coder %d: uncoded symbol %d (0x%x)\n",id,symbol,symbol);
#endif
		return;
		}
	if ((cnt = enclength[symbol]) != 0)
		pfio->putbits(encode[symbol],cnt);
#ifdef DEBUG
	if (debug)
		printf("Coder %d: symbol %d (0x%x)\n",id,symbol,symbol);
#endif
	if (longest && !cnt)
		printf("Tried to code untrained symbol %d\n",symbol);
	//printf("unenc = %d, coded %d bits\n",symbol,cnt);
}

// Member functions for the rlcomp type implement a "flagword" Huffman
// coder where, if the most frequent character is frequent enough, the
// patterns of freqchar/nonfreqchar are represented in a separate
// coder of flag words.

rlcomp::rlcomp(){
		debug = 0;
		id = 0;
		inbuf = 0;
		procbuf = 0;
		symbolcnt = 0;
		psymbols = new compactor;
		pflagwords = new compactor;
		reset();
		}
rlcomp::~rlcomp() {
		delete [] inbuf;
		delete [] procbuf;
		delete psymbols;
		delete pflagwords;
		}
int rlcomp::encodelength(int maxlength){ // Called to set up encode buffers
		buflength = maxlength;
		if (!maxlength){	// This is a decoder - no buffer
			inbuf = 0;
			procbuf = 0;
			return 0;
			}
		inbuf = new unsigned char [maxlength+FLAGWORDLENGTH];
		procbuf = new unsigned char [(maxlength/FLAGWORDLENGTH)+1];
		if (!inbuf || !procbuf)
			return -1;
		inbufend = inbuf;
		procbufend = procbuf;
		readproc = procbuf;
		return 1;
		}
void rlcomp::reset() {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		for (int i = 0; i < 256; i++)
			freq[i] = 0;
		symbolcnt = 0;
		doublecode = 0;
		inbufend = inbuf; 
		procbufend = procbuf;
		readproc = procbuf;
		usingflagwords = 0; // Not really necessary. Set when building
					// or reading tree.
		psymbols->reset();
		pflagwords->reset();
#ifdef NOCOMPACTORDEBUG
			enable_compactor_debug();
#endif
		}
void rlcomp::buildtree(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		// Does major part of flagword coding
		unsigned char *readin = inbuf;
		unsigned char *writeproc = procbuf;
		unsigned char symbol;
// Debug
//		printf("inbuf:\n");
//		while(readin < inbufend)
//			printf("%x ", (int) *readin++);
//		printf("\n");
//		readin = inbuf;
// End debug
// First find the most frequent symbol
		int maxfreq = -1;
		int i;
		for (i = 0; i < 256; i++) {
			if (freq[i] > maxfreq) {
				maxfreq = freq[i];
				freqsymbol = i;
				}
			}
// Now see if it is frequent enough to justify using flagwords
// See Cheung & Keily, "An Efficient Variable Length Coding Scheme for an
// IID Source".
		if ((maxfreq < 16)||(maxfreq*2 < inbufend-inbuf)||
			(maxfreq == inbufend-inbuf)){	
			// Too few to bother or 
			// Less than half are freqsymbol or
			// Single symbol
#ifndef NOCOMPACTORDEBUG
			if (debug)
				printf("rlcoder %d: Not using flagwords because maxfreq = %d and numchars = %d\n",id,maxfreq,(int)(inbufend-inbuf));
#endif
			usingflagwords = 0;
			// Set up symbol coder
			while(readin < inbufend)
				psymbols->inc(*readin++);
			// Don't do anything with flagword coder
			psymbols->buildtree();
			inbufend = inbuf;	// For consistency. Not necess.
#ifdef NOCOMPACTORDEBUG
			enable_compactor_debug();
#endif
			return;
			}
		usingflagwords = 1;
#ifndef NOCOMPACTORDEBUG
		if (debug)
			printf("rlcoder %d: Using flagwords because maxfreq = %d and numchars = %d\n",id,maxfreq,(int)(inbufend-inbuf));
#endif
		int numproc = (int)(inbufend-inbuf+(FLAGWORDLENGTH-1))/FLAGWORDLENGTH;
		unsigned char *padinbufend = inbuf;
		padinbufend += (numproc*FLAGWORDLENGTH);
		for (unsigned char *p = inbufend; p < padinbufend; p++)
			*p = freqsymbol;
		int allfreq = (1<<FLAGWORDLENGTH) -1;
		for (i = 0; i < numproc; i++) {
			unsigned char outbyte = 0;
			for (int j = 0; j < FLAGWORDLENGTH; j++) {
				outbyte <<= 1;
				if ((symbol = *readin++) == freqsymbol)
					outbyte |= 0x1;
				else
					psymbols->inc(symbol);
				}
			if ((outbyte == allfreq)&&(writeproc > procbuf)&&
				(*(writeproc-1) == allfreq)) {
			// Check for two successive all frequent symbols
			// Fold into one
				*(writeproc-1) = outbyte+1; // Repeat word
				pflagwords->dec(outbyte);
				pflagwords->inc(outbyte+1);
				}
			else {
				*writeproc++ = outbyte;
				pflagwords->inc(outbyte);
				}
			}
		procbufend = writeproc;
// Debug
//		writeproc = procbuf;
//		printf("procbuf:\n");
//		while(writeproc < procbufend)
//			printf("%x ", (int) *writeproc++);
//		printf("\n");
// End debug
		psymbols->buildtree();
		pflagwords->buildtree();
		inbufend = inbuf; // Now to be used for debugging
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::writetree(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		if (usingflagwords){
			putbit(1);
			int temp = (int) freqsymbol;
			for (int i = 7; i >= 0; i--) {
				putbit((temp>>i) & 1);
				}
			}
		else
			putbit(0);
		psymbols->writetree(); 
		if (usingflagwords)
			pflagwords->writetree();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::readtree(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		if (getbit()) {
			usingflagwords = 1;
			int temp = 0;
			for (int i = 0; i < 8; i++) {
				temp += getbit();
				temp <<= 1;
				}
			temp >>= 1;
			freqsymbol = (unsigned char) temp;
#ifndef NOCOMPACTORDEBUG
			if (debug)
				printf("rlcoder %d: Using flagwords, freqsymbol = %x\n",id,freqsymbol);
#endif
			symbolcnt = 0;
			doublecode = 0;
			}
		else {
#ifndef NOCOMPACTORDEBUG
			if (debug)
				printf("rlcoder %d: Not using flagwords\n",id);
#endif
			usingflagwords = 0;
			}
		psymbols->readtree(); 
		if (usingflagwords)
			pflagwords->readtree();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::maketable_encode(void){
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		psymbols->maketable_encode(); 
		if (usingflagwords)
			pflagwords->maketable_encode();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::maketable_decode(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		psymbols->maketable_decode(); 
		if (usingflagwords)
			pflagwords->maketable_decode();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::incorput(int symbol, int which) {
		if (which)
			put(symbol);
		else
			inc(symbol);
		}
void rlcomp::put(int inbyte) {
		if (!usingflagwords) {
			psymbols->put(inbyte);
			return;
			}
		if (inbyte != *inbufend++) {
			fprintf(stderr,"Internal error in rlcomp\n");
			// exit(1);
			}
		if (!(symbolcnt%FLAGWORDLENGTH)) {
			if (doublecode)
				doublecode--;
			else {
				if (*readproc == (1<<FLAGWORDLENGTH))
					doublecode = 1;
				pflagwords->put(*readproc++);
				}
			}
		if (inbyte != freqsymbol)
			psymbols->put(inbyte);
		symbolcnt++;
		}
