//
// Binary tree predictive decoder
//
// Version 5.
//
// Copyright (c) John Robinson. 1994, 1995, 1997, 2002, 2003.
//
// Converts BTPC to PGM or PPM using btpc_decode() function
//
#include <stdlib.h>
#include <string.h>
#include "btpc.h"

int writepnm(const char *pgmname, const unsigned char ***pbuf, 
const int& rows, const int& cols, const int& maxval, const int& pnmtype)
// Write picture to "P5" pgm or "P6" ppm format
// (If a "P4" format pbm picture was encoded, btpc converts it into a
// two-level pgm picture.)
	{
	FILE *out = fopen(pgmname, "wb");
	if (!out) {
		fprintf(stderr,"Cannot open %s\n", pgmname);
		exit (0);
		}
	if (pnmtype == 1) {	// PGM
	    fprintf(out,"P5\n%d %d\n%d\n", cols, rows, maxval);
	    for (int i = 0; i < rows; i++)
		fwrite(pbuf[0][i],1,cols,out);
	    }
	else { // PPM
	    fprintf(out,"P6\n%d %d\n%d\n", cols, rows, maxval);
	    unsigned char *temp = new unsigned char[3*cols];
	    for (int i = 0; i < rows; i++) {
		register unsigned char *ptemp = temp;
		register const unsigned char *pg = pbuf[0][i];
		register const unsigned char *pr = pbuf[1][i];
		register const unsigned char *pb = pbuf[2][i];
		for (int j = 0; j < cols; j++) {
			*ptemp++ = *pr++;
			*ptemp++ = *pg++;
			*ptemp++ = *pb++;
			}
		fwrite(temp,1,cols*3,out);
		}
	    delete [] temp;
	    }
	fclose(out);
	return 0;
	}

int main(int argc, const char *argv[])
	{
	if (argc != 3) {
		fprintf(stderr, "Usage: dbtpc code_file pgm_outpic\n");
		return -1;
		}
	unsigned char **components[3];
	char *comment;
	int truerows, truecols;
	int pnmtype;
	int maxval;
	int retval;
	FILE *cfp = fopen(argv[1],"rb");
	if (!cfp) 
		retval = -1;
	else
		retval = btpc_decode(components, cfp,
			truerows, truecols, pnmtype, maxval,&comment);
	switch (retval) {
		case -1:
			fprintf(stderr, "Can't open %s\n", argv[1]);
			return -1;
		case -2:
			fprintf(stderr, "%s is not a valid btpc5 file\n", argv[1]);
			return -2;	// Not a btpc file
		case -3:
			fprintf(stderr, "%s is coded with a later version of btpc5. Cannot decode.\n", argv[1]);
			return -3;	// Not a btpc file
		default:
			break;
		}
	writepnm(argv[2], (const unsigned char ***) components, 
		truerows, truecols, maxval,pnmtype);
	btpc_finish(components, truerows, truecols, pnmtype);
	if (*comment)
		printf("Comment: %s\n",comment);
	return retval;
	}
		

