//
// Binary tree predictive coding
// Helper routines for encoding, plus main encoding function
//
// Note: I have not optimized the encoding.
//
// Copyright (C) John Robinson, 1994, 1995, 1997, 2002, 2003.
//
// #define INCEDGE 1
#define ENCODER 1
#include <string.h>
#include "btpc.h"
#include "order.h"
// #define DEBUG 1
const int numlevels = 6;
int btpc_get_numlevels() {
	return numlevels;
	}

#ifdef DEBUG
//
// stuff for debugging.
//
int writepgm(const char *pgmname, unsigned char **pbuf,
const int& rows, const int& cols, const int& maxval)
// Write picture to "P5" pgm format
        {
        FILE *out = fopen(pgmname, "w");
        if (!out) {
                fprintf(stderr,"Cannot open %s\n", pgmname);
                exit (0);
                }
        fprintf(out,"P5\n%d %d\n%d\n", cols, rows, maxval);
        for (int i = 0; i < rows; i++)
                fwrite(pbuf[i],1,cols,out);
        fclose(out);
        return 0;
        }
//
// End of new stuff for debugging
//
#endif

void pad(unsigned char ***incomponents, unsigned char ***pbuf, 
	int& rows, int& cols, int pnmtype)
// Read picture in "P5" or "P6" format into arrays. Pad the picture size up to
// the next multiple of blocksize.
	{
	int i, j;
	// Make ymax equal to next highest multiple of blocksize
	int ymax = ((rows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	// Make xmax equal to next highest multiple of 2*blocksize
	// Do this so that points can be processed in pairs in binary tree
	int xmax = ((cols+(1<<(numlevels+1))-1)>>(numlevels+1))<<(numlevels+1);
	if (pnmtype == 1)	// PGM
	    {
	    unsigned char **buf;
	    *pbuf++ = buf = new unsigned char *[ymax];
	    *pbuf++ = buf;
	    *pbuf = buf;	// So all three components point at same
	    for (i = 0; i < rows; i++) {
		buf[i] = new unsigned char[xmax];
		memcpy(buf[i],incomponents[0][i],cols);
		for (j = cols; j < xmax; j++)	// Pad image by
			buf[i][j] = buf[i][cols-1];	// copying last pel
		}
	    // Vertical padding
	    for (i = rows; i < ymax; i++) {
		buf[i] = new unsigned char[xmax];	// Pad image by
		memcpy(buf[i], buf[rows-1], xmax);	// copying last line
		}
	    }
	else			// PPM
	    {
	    unsigned char **bufr, **bufg, **bufb;
	    *pbuf++ = bufg = new unsigned char *[ymax];
	    *pbuf++ = bufr = new unsigned char *[ymax];
	    *pbuf = bufb = new unsigned char *[ymax];
	    for (i = 0; i < rows; i++){
		bufg[i] = new unsigned char[xmax];
		bufr[i] = new unsigned char[xmax];
		bufb[i] = new unsigned char[xmax];
		memcpy(bufg[i],incomponents[0][i],cols);
		memcpy(bufr[i],incomponents[1][i],cols);
		memcpy(bufb[i],incomponents[2][i],cols);
		for (j = cols; j < xmax; j++) {	// Pad
			bufr[i][j] = bufr[i][cols-1];
			bufg[i][j] = bufg[i][cols-1];
			bufb[i][j] = bufb[i][cols-1];
			}
		}
	    // Vertical padding
	    for (i = rows; i < ymax; i++) {
		bufg[i] = new unsigned char[xmax];
		bufr[i] = new unsigned char[xmax];
		bufb[i] = new unsigned char[xmax];
		memcpy(bufg[i], bufg[rows-1], xmax);	// copying last line
		memcpy(bufr[i], bufr[rows-1], xmax);	// copying last line
		memcpy(bufb[i], bufb[rows-1], xmax);	// copying last line
		}
	    }
	}

static int Asq_encode_table[513] = { 0};
static int *sq_encode_table;
static int Asq_decode_table[513] = { 0};
static int *sq_decode_table;
static int sq_range;
static int sq_arr[19];  // Large enough for any number of levels we might use
static int Adi_encode_table[513] = { 0};
static int *di_encode_table;
static int Adi_decode_table[513] = { 0};
static int *di_decode_table;
static int di_range;
static int di_arr[19];  // Large enough for any number of levels we might use
static int Aabt[513];
static int *abt;

void prepare_encoder(const int& sq_quant_spacing, const int& di_quant_spacing,
	const int& maximum)
// Sets up tables of representative levels to use in the current bands.
{
    int i;
    abt = &Aabt[256];
    abt[0] = 0;
    for (i = 1; i < 256; i++) {
	abt[i] = i;
	abt[-i] = i;
	}
    // Following six lines required to initialize tables if btpc_encode()
    // used repeatedly in same program. Previously relied on the
    // static initializations above.
    for (i = 0; i < 513; i++) {
	Asq_encode_table[i] = 0;
	Adi_encode_table[i] = 0;
	Asq_decode_table[i] = 0;
	Adi_decode_table[i] = 0;
	}
    // First do the square band
    // printf("Square band quantization: spacing = %d\n",sq_quant_spacing);
    sq_range = maximum/sq_quant_spacing + 1;
    sq_encode_table = &Asq_encode_table[256];
    for (i = 1; i < 256; i++) {
	sq_encode_table[i] = i/sq_quant_spacing;
	sq_encode_table[-i] = -i/sq_quant_spacing;
	}
    sq_decode_table = &Asq_decode_table[256];
    int rep_level = sq_quant_spacing + (sq_quant_spacing-1)/2;
    // printf("Range = %d, first rep level = %d.\n",sq_range,rep_level);
    sq_decode_table[0] = 0;
    for (i = 1; i <= sq_range; i++, rep_level += sq_quant_spacing) {
	sq_decode_table[i] = rep_level;
	sq_decode_table[-i] = -rep_level;
    	}
    // Then the diamond band
    // printf("Diamond band quantization: spacing = %d\n",di_quant_spacing);
    di_range = maximum/di_quant_spacing + 1; 
    di_encode_table = &Adi_encode_table[256];
    for (i = 1; i < 256; i++) {
	di_encode_table[i] = i/di_quant_spacing;
	di_encode_table[-i] = -i/di_quant_spacing;
	}
    di_decode_table = &Adi_decode_table[256];
    rep_level = di_quant_spacing + (di_quant_spacing-1)/2;
    // printf("Range = %d, first rep level = %d.\n",di_range,rep_level);
    di_decode_table[0] = 0;
    for (i = 1; i <= di_range; i++, rep_level += di_quant_spacing) {
	di_decode_table[i] = rep_level;
	di_decode_table[-i] = -rep_level;
    	}
    }
static unsigned char **scratch;
static unsigned char **predictors;
static unsigned char **ridge;
#ifdef DEBUG
static unsigned char **debugimage;
#endif

int makescratch(const int& xmax, const int& ymax)
// Make space for scratch images. Assume we don't run out of space.
	{
	scratch = new unsigned char *[ymax];
	ridge = new unsigned char *[ymax];
#ifdef DEBUG
	debugimage = new unsigned char *[ymax];
#endif
	predictors = new unsigned char *[ymax];
	for (int i = 0; i < ymax; i++)
		{
		scratch[i] = new unsigned char[xmax];
		ridge[i] = new unsigned char[xmax];
		memset(ridge[i], '\0', xmax);
#ifdef DEBUG
		debugimage[i] = new unsigned char[xmax];
		memset(debugimage[i], '\0', xmax);
#endif
		predictors[i] = new unsigned char[xmax];
		memset(predictors[i], 255, xmax);
		}
	return 0;
	}
int deletescratch(const int& ymax)
	{
	for (int i = 0; i < ymax; i++) {
	    delete [] scratch[i];
	    delete [] ridge[i];
#ifdef DEBUG
	    delete [] debugimage[i];
#endif
	    delete [] predictors[i];
	    }
	delete [] scratch;
	delete [] ridge;
#ifdef DEBUG
	delete [] debugimage;
#endif
	delete [] predictors;
	return 0;
	}

// Calculate quantization levels from quality parameter
int calcq(int dynamicrange, int quality, int *qvals, int numbands) {
	if ((quality > 100) || (quality < 0))
		return -1;
	int topval = (100-quality)/2;
	// topval will be the quantizer spacing on the finest band
	topval *= dynamicrange;
	topval /= 255;
	if (!topval) {	// Quality == 100 or scaling makes it so
		for (int i = 0; i < numbands; i++)
			*qvals++ = 1;
		return 1;
		}
	double qratio;
	if (topval >= 40)
		{
		topval += (topval-25) + 3*(topval-40);
		// qratio = 0.75;
		qratio = 0.80;
		}
	else if (topval > 25)
		{
		topval += (topval-25);
		// qratio = 0.75;
		qratio = 0.80;
		}
	else
		qratio = 0.8;
		// qratio = 0.85;
	qvals[numbands-1] = topval;
	int i;
	for (i = numbands-2; i >= 0; i--)
		// qvals[i] = (int)((((double)qvals[i+1])*qratio)+0.99999);
		qvals[i] = (int)((((double)qvals[i+1])*qratio)+0.6666);
#ifdef DEBUG
	printf("Quality parameter = %d, dynamic range = %d.\n", 
		quality, dynamicrange);
	printf("Quantizater spacings: ");
	for (i = 0; i < numbands; i++) {
		printf("%d ",qvals[i]);
		}
	printf("\n");
#endif
	return 1;
	}

void gettrueminmax(unsigned char ***components, const int pnmtype,
		int *minval, int *maxval, const int& rows, const int& cols){
	unsigned char *p;
	for (int comp = 0; comp < pnmtype; comp++) {
		int minv = 256;
		int maxv = -1;
		for (int i = 0; i < rows; i++) {
			p = components[comp][i];
			for (int j = 0; j < cols; j++) {
				if (*p < minv) minv = *p;
				if (*p > maxv) maxv = *p;
				p++;
				}
			}
		*minval++ = minv;
		*maxval++ = maxv;
		}
	}
void from_bilevel(unsigned char ***components, const int pnmtype,
		int *minval, int *maxval, const int& rows, const int& cols,
		int &bilevel) {
	// do work to check if only two values
	bilevel = 0;
	unsigned char *p;
	int firstval = minval[0];
	int secondval = -1;
	int comp;
	for (comp = 0; comp < pnmtype; comp++) {
		for (int i = 0; i < rows; i++) {
			p = components[comp][i];
			for (int j = 0; j < cols; j++) {
				if (secondval < 0) {
					if (*p != firstval)
						secondval = *p;
					}
				else {
					if ((*p != firstval)&&
					    (*p != secondval)) {
						return;
						}
					}
				p++;
				}
			}
		}
	bilevel = 1; // Got to here so it is bilevel
	if (secondval < firstval){ // Conceivable if first comp is flat and
				   // second is that value and one lower
		firstval = secondval;
		secondval = minval[0];
		}
	// Make the transformation
	for (comp = 0; comp < pnmtype; comp++) {
		for (int i = 0; i < rows; i++) {
			p = components[comp][i];
			for (int j = 0; j < cols; j++) {
				if (*p == firstval)
					*p = 0;
				else
					*p = 1;
				p++;
				}
			}
		}
	}



void btpc_enc_level(unsigned char **in,
const int& XMAX, const int& YMAX, const int& level, 
const int& sq_spacing, const int& di_spacing, 
const int& minval, const int& maxval, const int& comp, 
const int& truecols, const int& truerows) {
    register int error;
    int code;
    register int prepel;
    int X, Y;
    int ridgepel;
    register int p,q,r,s, t, u, v;	
    // Surround pels
    register int estimate;
    register int elow, ehigh;
    prepare_encoder(sq_spacing, di_spacing, maxval);
    sq_arr[level] = sq_range;
    di_arr[level] = di_range;

    int sq_ridgethresh = btpc_get_ridgethresh(sq_spacing);
    if (maxval < sq_ridgethresh)	// Use maxval rather than maxval-minval
	sq_ridgethresh = maxval;	// because range is just based on max
    int di_ridgethresh = btpc_get_ridgethresh(di_spacing);
    if (maxval < di_ridgethresh)
	di_ridgethresh = maxval;
    const int XINTERVAL = 1 << level;
    const int YINTERVAL = 1 << level;
    const int XHINT = XINTERVAL >> 1;
    const int YHINT = YINTERVAL >> 1;
#ifdef DEBUG
    // printf("level = %d\n",level);
#endif
    for (Y = YHINT; Y < YMAX; Y += YINTERVAL)
    {
#ifdef DEBUG
	// printf("Y = %d\n",Y); fflush(stdout);
#endif
	int yplus = Y+YHINT;
	int yminus = Y-YHINT;
	int y2minus = Y-YINTERVAL;
	if (y2minus < 0)
		y2minus = 0;
	int y3minus = y2minus-YHINT;
	if (y3minus < 0)
		y3minus = 0;
	unsigned char *py = in[Y];
	unsigned char *pyplus;
	if (yplus == YMAX)
		pyplus = in[yminus]; // So at bottom of picture, take from
				     // previous line
	else
		pyplus = in[yplus];
	unsigned char *pyminus = in[yminus];
	unsigned char *py2minus = in[y2minus];
	unsigned char *py3minus = in[y3minus];
	// Set up q and r and v for first row
	v = q = pyminus[0];
	r = pyplus[0];
	for (X = XHINT; X < XMAX; ) {
#ifdef DEBUG
	    // printf("X = %d\n",X); fflush(stdout);
#endif
	    int xplus = X+XHINT;
	    int xminus = X-XHINT;
	    p = q;
	    s = r;
	    u = v;
	    if (xplus != XMAX) {
		q = pyminus[xplus];
		r = pyplus[xplus];
	    }
	    // Otherwise q and r are already correct
	    if (!yminus)
		t = p;
	    else
		t = py2minus[X];
	    prepel = py[X];
	    int predictor;
	    predictor = predictors[Y][X];
	    ridgepel = ridge[Y][X];
#define SQUARE 1
#include "macro.cpp"
#undef SQUARE
	    py[X] = prepel;
	    scratch[Y][X] = code;
	    ridge[Y][X] = ridgepel;
#ifdef DEBUG
	//      debugimage[Y][X] = prepel - estimate;
		debugimage[Y][X] = predictor;
	//	if (ridgepel)
	//		printf("ridgepel = %d at %d,%d\n",ridgepel,X,Y);
	//	printf("check A\n"); fflush(stdout);
#endif
	    predictors[Y][X] = predictor;

	    v = prepel;
	    prepel = pyminus[X];
	    predictor = predictors[yminus][X];
	    ridgepel = ridge[yminus][X];
#define DIAMOND_LEFT 1
#include "macro.cpp"
#undef DIAMOND_LEFT
	    // printf("check B\n"); fflush(stdout);
	    pyminus[X] = prepel;
	    scratch[yminus][X] = code;
	    ridge[yminus][X] = ridgepel;
#ifdef DEBUG
	//    	debugimage[yminus][X] = prepel - estimate;
		debugimage[yminus][X] = predictor;
	//	if (ridgepel)
	//		printf("ridgepel = %d at %d,%d\n",ridgepel,X,yminus);
#endif
	    predictors[yminus][X] = predictor;

	    prepel = py[xminus];
	    predictor = predictors[Y][xminus];
	    ridgepel = ridge[Y][xminus];
#define DIAMOND_RIGHT 1
#include "macro.cpp"
#undef DIAMOND_RIGHT
	    py[xminus] = prepel;
	    scratch[Y][xminus] = code;
	    ridge[Y][xminus] = ridgepel;
#ifdef DEBUG
	//	debugimage[Y][xminus] = prepel - estimate;
		debugimage[Y][xminus] = predictor;
	//	if (ridgepel)
	//		printf("ridgepel = %d at %d,%d\n",ridgepel,xminus,Y);
	//	printf("check C\n"); fflush(stdout);
#endif
	    predictors[Y][xminus] = predictor;

	    X += XINTERVAL;
	    }
	}
    }
// See comments in the middle of bincode for how this converts internal
// representation in scratch array into a coder index and topcodeval.
void mktopcode( const unsigned char scratchval,
    const unsigned char& l0, const unsigned char& l1, const unsigned char& l2,
    unsigned char rscratchval,	// Not const so can force to zero
    const unsigned char& r0, const unsigned char& r1, const unsigned char& r2,
    const int pnmtype, rlcomp *codeout, const int putflag, 
    const int sendright, const int& range) {
	unsigned char send0, send1, send2, rsend0, rsend1, rsend2;
	int coder;
	unsigned char topcodeval;
	send0 = l0;
	rsend0 = r0;
	if (pnmtype == 1) {
		rsend1 = rsend2 = send1 = send2 = 0;
		}
	else {
		send1 = l1;
		send2 = l2;
		rsend1 = r1;
		rsend2 = r2;
		}
	if (!sendright) {
		rsend0 = rsend1 = rsend2 = 0;
		rscratchval = 0;
		}
	coder = (scratchval & 3) ? 1 : 0;
	coder |= (scratchval & 0xc) ? 2 : 0;
	coder |= (scratchval & 0x30) ? 4 : 0;
	if (coder == 0)
		return;
	if (pnmtype != 1) {
		unsigned char nextcoderleft, nextcoderright;
		nextcoderleft = 0;
		nextcoderleft |= (scratchval & 2) ? 1 : 0;
		nextcoderleft |= (scratchval & 0x8) ? 2 : 0;
		nextcoderleft |= (scratchval & 0x20) ? 4 : 0;
		nextcoderright = 0;
		nextcoderright |= (rscratchval & 2) ? 1 : 0;
		nextcoderright |= (rscratchval & 0x8) ? 2 : 0;
		nextcoderright |= (rscratchval & 0x20) ? 4 : 0;
		topcodeval = 0;
		if (nextcoderleft)
			topcodeval |= 0x80;
		if (nextcoderright)
			topcodeval |= 0x40;
		if (send0)
			topcodeval |= 0x8;
		if (send1)
			topcodeval |= 0x10;
		if (send2)
			topcodeval |= 0x20;
		if (rsend0)
			topcodeval |= 0x1;
		if (rsend1)
			topcodeval |= 0x2;
		if (rsend2)
			topcodeval |= 0x4;
		}
	else {
		topcodeval = 0;
		topcodeval |= (scratchval & 0x2) ? 2 : 0;
		topcodeval |= (rscratchval & 0x2) ? 1 : 0;
		topcodeval <<= 4;
		if (send0) {
		    if (send0 == 1)
			{ topcodeval |= 0x4; send0 = 0; }
		    else if (send0 == range-1)
			{ topcodeval |= 0x8; send0 = 0; }
		    else
			{ topcodeval |= 0xc; }
		    }
		if (rsend0) {
		    if (rsend0 == 1)
			{ topcodeval |= 0x1; rsend0 = 0; }
		    else if (rsend0 == range-1)
			{ topcodeval |= 0x2; rsend0 = 0; }
		    else
			{ topcodeval |= 0x3; }
		    }
		}
	codeout[3].incorput(topcodeval,putflag);
	if (send0){
		if (send0 >= range-2) {
			send0 = range-send0;
			codeout[0].incorput(send0,putflag);
			if (putflag)
				codeout[0].putbit(1);
			}
		else if (send0 <= 2) {
			codeout[0].incorput(send0,putflag);
			if (putflag)
				codeout[0].putbit(0);
			}
		else
			codeout[0].incorput(send0,putflag);
		}
	if (send1){
		if (send1 >= range-2) {
			send1 = range-send1;
			codeout[1].incorput(send1,putflag);
			if (putflag)
				codeout[0].putbit(1);
			}
		else if (send1 <= 2) {
			codeout[1].incorput(send1,putflag);
			if (putflag)
				codeout[0].putbit(0);
			}
		else
			codeout[1].incorput(send1,putflag);
		}
	if (send2){
		if (send2 >= range-2) {
			send2 = range-send2;
			codeout[2].incorput(send2,putflag);
			if (putflag)
				codeout[0].putbit(1);
			}
		else if (send2 <= 2) {
			codeout[2].incorput(send2,putflag);
			if (putflag)
				codeout[0].putbit(0);
			}
		else
			codeout[2].incorput(send2,putflag);
		}
	if (!sendright)
		return;
	if (rsend0){
		if (rsend0 >= range-2) {
			rsend0 = range-rsend0;
			codeout[0].incorput(rsend0,putflag);
			if (putflag)
				codeout[0].putbit(1);
			}
		else if (rsend0 <= 2) {
			codeout[0].incorput(rsend0,putflag);
			if (putflag)
				codeout[0].putbit(0);
			}
		else
			codeout[0].incorput(rsend0,putflag);
		}
	if (rsend1){
		if (rsend1 >= range-2) {
			rsend1 = range-rsend1;
			codeout[1].incorput(rsend1,putflag);
			if (putflag)
				codeout[0].putbit(1);
			}
		else if (rsend1 <= 2) {
			codeout[1].incorput(rsend1,putflag);
			if (putflag)
				codeout[0].putbit(0);
			}
		else
			codeout[1].incorput(rsend1,putflag);
		}
	if (rsend2){
		if (rsend2 >= range-2) {
			rsend2 = range-rsend2;
			codeout[2].incorput(rsend2,putflag);
			if (putflag)
				codeout[0].putbit(1);
			}
		else if (rsend2 <= 2) {
			codeout[2].incorput(rsend2,putflag);
			if (putflag)
				codeout[0].putbit(0);
			}
		else
			codeout[2].incorput(rsend2,putflag);
		}
	}
	
//
// bincode works from the finest band upwards, chunking zeros up the binary
// tree until a non-zero entry is reached, which is marked as a leaf. It
// then works down the tree, Huffman coding each band.
//
int bincode(unsigned char ***pyr, 
	const int& inrows, const int& incols, 
	rlcomp *codeout, 
	int fromlevel, const int *spacing, const int pnmtype, const int& maxval)
        {
	int x, y;
	int i,j;
	int rpel;
	int scratchpel;
	// Deal with DPCM coding of coarsest band first
	for (int comp = 0; comp < pnmtype; comp++) {
	    int predpel;
	    int lpspacing = 1<<fromlevel;
	    for (y = 0; y < inrows; y += lpspacing)
	    for (x = 0; x < incols; x += lpspacing)
		{
		predpel =  (x ? pyr[comp][y][x-lpspacing] : (y ? pyr[comp][y-lpspacing][x] : 0));
		int temp = pyr[comp][y][x] - predpel;
		// Ensure wrapping round
		if (temp < -128) temp += 256;
		if (temp > 127) temp -= 256;
		codeout[0].inc(scratch[y][x] = (unsigned char) temp);
		}
	    codeout[0].buildtree();
	    codeout[0].writetree();
	    for (y = 0; y < inrows; y += lpspacing)
	    for (x = 0; x < incols; x += lpspacing){
		codeout[0].put((unsigned char) scratch[y][x]); 
#ifdef DEBUG
		scratch[y][x] = 0;
		pyr[comp][y][x] = 0;
#endif
		}
	    codeout[0].reset();
	    }
	// Flags on scratch image interpreted as follows:
	// 0: Beyond end of a tree - don't code
	// 1: Tree terminator
	// 2: Ordinary
	// Bottom 2 bits channel 0, next 2 bits channel 1, next 2 bits chan 2
	const int rows = inrows;
	const int cols = incols;
	int fillbyte = 0x2a;	// Binary 00101010
	if (pnmtype == 1)
		fillbyte = 2;
	for (y = 0; y < rows; y++)
		memset(scratch[y],fillbyte,cols);
	// First blank out highest pass
	int yodd = 0;
	for (y = 0; y < rows; y++, yodd ^= 1)
	    {
	    for (x = (yodd? 0:1); x < cols; x += 2)
		{
		scratchpel = scratch[y][x];
		if (pyr[0][y][x] == 0) 
			scratchpel = (scratchpel & 0x3c) | 1;
		if ((pnmtype > 1)&&(pyr[1][y][x] == 0))
			scratchpel = (scratchpel & 0x33) | 4;
		if ((pnmtype > 1)&&(pyr[2][y][x] == 0))
			scratchpel = (scratchpel & 0x0f) | 0x10;
		scratch[y][x] = scratchpel;
		}
	    }
	// Now first mpass
	for (y = 1; y < rows; y += 2)
	for (x = 1; x < cols; x += 2)
		{
		if (((scratch[y-1][x] & 0x03) == 1)&&
		    ((scratch[y][x-1] & 0x03) == 1)) {
			scratch[y][x] = (scratch[y][x] & 0x3c) | 1;
			scratch[y-1][x] = scratch[y-1][x] & 0x3c;
			scratch[y][x-1] = scratch[y][x-1] & 0x3c;
			}
		if (pnmtype == 1)
			continue;
		if (((scratch[y-1][x] & 0x0c) == 4)&&
		    ((scratch[y][x-1] & 0x0c) == 4)) {
			scratch[y][x] = (scratch[y][x] & 0x33) | 4;
			scratch[y-1][x] = scratch[y-1][x] & 0x33;
			scratch[y][x-1] = scratch[y][x-1] & 0x33;
			}
		if (((scratch[y-1][x] & 0x30) == 0x10)&&
		    ((scratch[y][x-1] & 0x30) == 0x10)) {
			scratch[y][x] = (scratch[y][x] & 0x0f) | 0x10;
			scratch[y-1][x] = scratch[y-1][x] & 0x0f;
			scratch[y][x-1] = scratch[y][x-1] & 0x0f;
			}
		}
	// Now remainder of bands
	for (i = 2; i <= 2*fromlevel; i *= 2)
		{
		int twoi = i<<1;
		int halfi = i>>1;
		yodd = 0;
		for (y = 0; y < rows; y += i, yodd ^= 1)
		    {
		    int lchild = yodd? halfi : -halfi;
		    for (x = (yodd? 0: i); x < cols; x += twoi)
			  {
			  if (((scratch[y+halfi][x+lchild] & 0x03) == 1)&&
			    (pyr[0][y+halfi][x+lchild] == 0)&&
			    ((scratch[y+halfi][x+lchild+i] & 0x03) == 1)&&
			    (pyr[0][y+halfi][x+lchild+i] == 0)) {
				scratch[y][x]  = (scratch[y][x] & 0x3c) | 1;
			  	scratch[y+halfi][x+lchild] &= 0x3c;
			   	scratch[y+halfi][x+lchild+i] &= 0x3c;
				}
			    if (pnmtype == 1)
				continue;
			  if (((scratch[y+halfi][x+lchild] & 0x0c) == 4)&&
			    (pyr[1][y+halfi][x+lchild] == 0)&&
			    ((scratch[y+halfi][x+lchild+i] & 0x0c) == 4)&&
			    (pyr[1][y+halfi][x+lchild+i] == 0)) {
				scratch[y][x]  = (scratch[y][x] & 0x33) | 4;
			  	scratch[y+halfi][x+lchild] &= 0x33;
			   	scratch[y+halfi][x+lchild+i] &= 0x33;
				}
			  if (((scratch[y+halfi][x+lchild] & 0x30) == 0x10)&&
			    (pyr[2][y+halfi][x+lchild] == 0)&&
			    ((scratch[y+halfi][x+lchild+i] & 0x30) == 0x10)&&
			    (pyr[2][y+halfi][x+lchild+i] == 0)) {
				scratch[y][x]  = (scratch[y][x] & 0x0f) | 0x10;
			  	scratch[y+halfi][x+lchild] &= 0x0f;
			   	scratch[y+halfi][x+lchild+i] &= 0x0f;
				}
			  }
		    }
		for (y = i; y < rows; y += twoi)
		for (x = i; x < cols; x += twoi)
			{
			if(((scratch[y-i][x] & 0x03) == 1)&&
			    (pyr[0][y-i][x] == 0)&&
			    ((scratch[y][x-i] & 0x03) == 1)&&
			    (pyr[0][y][x-i] == 0)) {
				scratch[y][x] = (scratch[y][x] & 0x3c) | 1;
			  	scratch[y-i][x] &= 0x3c;
			    	scratch[y][x-i] &= 0x3c;
				}
			if (pnmtype == 1)
				continue;
			if(((scratch[y-i][x] & 0x0c) == 4)&&
			    (pyr[1][y-i][x] == 0)&&
			    ((scratch[y][x-i] & 0x0c) == 4)&&
			    (pyr[1][y][x-i] == 0)) {
				scratch[y][x] = (scratch[y][x] & 0x33) | 4;
			  	scratch[y-i][x] &= 0x33;
			    	scratch[y][x-i] &= 0x33;
				}
			if(((scratch[y-i][x] & 0x30) == 0x10)&&
			    (pyr[2][y-i][x] == 0)&&
			    ((scratch[y][x-i] & 0x30) == 0x10)&&
			    (pyr[2][y][x-i] == 0)) {
				scratch[y][x] = (scratch[y][x] & 0x0f) | 0x10;
			  	scratch[y-i][x] &= 0x0f;
			    	scratch[y][x-i] &= 0x0f;
				}
			}
		}
	// Finally, change all the terminators in the highest band back
	// to zeros
	yodd = 0;
	for (y = 0; y < rows; y++, yodd ^= 1)
	    {
	    for (x = (yodd? 0:1); x < cols; x += 2) {
		scratchpel = scratch[y][x];
		if (scratchpel & 1)
			scratchpel ^= 3; // I.e. back to 2: ordinary
		if (scratchpel & 4)
			scratchpel ^= 0xc;
		if (scratchpel & 0x10)
			scratchpel ^= 0x30;
		scratch[y][x] = scratchpel;
		}
	    }

// Now interpret the codes and send them in the following way:
// From the scratch image determine which coders are still running and
// convert this to a 3-bit number which will index the array of codeout 
// objects. Then form a "topcode" output word where the bits have the
// meanings: bit 0: if 1, have to send children in channel 0; if 0 this is
// either a tree termination or below a tree termination. bit 1: if 1, have
// to send children in channel 1; bit 2: if 2, have to send children in channel
// 2. Therefore the bit pattern in the lowest three bits corresponds to the
// coder index for the following band. bit 4: if 1, channel 0 value is non-zero.
// bit 5: if 1, channel 1 value is non-zero. bit 6; if 1, channel 2 value
// is non-zero.
	int band = 1;
	for (i = fromlevel; i > 0; i--, band += 2)
		{
#ifdef DEBUG
		printf("LEVEL %d, band %d\n", i, band);
#endif
    		prepare_encoder(spacing[band], spacing[band+1], maxval);
		int hoffset = 1<<(i-1);
		int offset = 1<<i;
		for (int incorput = 0; incorput <= 1; incorput++) {
			for (y = hoffset; y < rows; y += offset)
			    {
#ifdef DEBUG
			    if (incorput)
		    		printf("Y = %d\n",y);
#endif
			    for (x = hoffset; x < cols; x += 2*offset)
				{
#ifdef DEBUG
				if (incorput)
					printf("X = %d\n",x);
#endif
				mktopcode(scratch[y][x], 
					pyr[0][y][x], pyr[1][y][x],
					pyr[2][y][x],
					scratch[y][x+offset], 
					pyr[0][y][x+offset], 
					pyr[1][y][x+offset], 
					pyr[2][y][x+offset], 
					pnmtype, codeout+SST, 
					incorput, (x + offset < cols),
					sq_arr[i]);
				if (incorput) {
					if ((rpel = ridge[y][x]) != 0)
						codeout[0].putbit(rpel - 2);
					// Defer flag for second pel because
					// decoder doesn't get to it yet.
					}
				mktopcode(scratch[y-hoffset][x], 
					pyr[0][y-hoffset][x], 
					pyr[1][y-hoffset][x],
					pyr[2][y-hoffset][x],
					scratch[y][x-hoffset], 
					pyr[0][y][x-hoffset], 
					pyr[1][y][x-hoffset], 
					pyr[2][y][x-hoffset], 
					pnmtype, codeout+DST, 
					incorput, 1,
					di_arr[i]);
				if (incorput) {
					if ((rpel = ridge[y-hoffset][x]) != 0)
						codeout[0].putbit(rpel - 2);
					if ((rpel = ridge[y][x-hoffset]) != 0)
						codeout[0].putbit(rpel - 2);
					// Now second ridge flag for square band
					if (x + offset < cols) {
					    if ((rpel = ridge[y][x+offset])!=0)
						codeout[0].putbit(rpel - 2);
						}
					}
				// if (x + offset >= cols)
				// 	break;
				mktopcode(scratch[y-hoffset][x+offset], 
					pyr[0][y-hoffset][x+offset], 
					pyr[1][y-hoffset][x+offset],
					pyr[2][y-hoffset][x+offset],
					scratch[y][x+hoffset], 
					pyr[0][y][x+hoffset], 
					pyr[1][y][x+hoffset], 
					pyr[2][y][x+hoffset], 
					pnmtype, codeout+DST, 
					incorput, 1,
					di_arr[i]);
				if (incorput) {
					if ((rpel = ridge[y-hoffset][x+offset]) != 0)
						codeout[0].putbit(rpel - 2);
					if ((rpel = ridge[y][x+hoffset]) != 0)
						codeout[0].putbit(rpel - 2);
					}
				}
		    	    }
			if (!incorput) {	// First time only
				for (j = 0; j < numcoders; j++) {
					codeout[j].buildtree();
					codeout[j].writetree();
					}
				}
			}
		for (j = 0; j < numcoders; j++) 
			codeout[j].reset();
		}
	return 0;
	}

void btpc_enc_plane(unsigned char **in, const int& xmax, const int& ymax,
const int& fromlevel, const int& tolevel, const int *spacing, 
rlcomp *codeout, const int& minval, const int& maxval, const int& comp,
const int& truecols, const int& truerows) {
#ifdef DEBUG
	static char planefile[7] = {'p','l','a','n','e','0','\0'};
	static char recfile[7] = {'r','e','c','o','v','0','\0'};
#endif
	// First put lowband into scratch
	int lpspacing = 1<<fromlevel;
	int y;
	for (y = 0; y < ymax; y += lpspacing)
	for (int x = 0; x < xmax; x += lpspacing)
		scratch[y][x] = in[y][x];
	// Do bands in pairs
	int band = 1;
	for (int level = fromlevel; level > tolevel; level--, band += 2)
		{
		btpc_enc_level(in,xmax,ymax,level,spacing[band],
		    spacing[band+1], minval, maxval, comp, 
		    truecols, truerows);
		}
#ifdef DEBUG
	writepgm(recfile,in,ymax,xmax,255);
#endif
	// Now copy encoded version back into in
	unsigned char *pfrom, *pto;
	for (y = 0; y < ymax; y++) {
		pto = in[y];
		pfrom = scratch[y];
		for (int x = 0; x < xmax; x++)
			*pto++ = *pfrom++;
		}
	// bincode(pyr, ymax, xmax, codeout, in, fromlevel, tolevel,
	// 		spacing);
#ifdef DEBUG
	writepgm(planefile,debugimage,ymax,xmax,255);
	++(planefile[5]);
	++(recfile[5]);
#endif
	}

int btpc_encode(unsigned char ***incomponents, FILE *cfp,
    int truerows, int truecols, int padrows, int padcols, int pnmtype,
    int quality, const char *comment) {
/*
	printf("Truerows = %d, truecols = %d, padrows = %d, padcols = %d\n",
		truerows, truecols, padrows, padcols);
*/
	unsigned char **components[3];
	int maxval[3], minval[3];
	int tmaxval[3], tminval[3];	// max and min of transformed versions
	int omaxval, ominval;		// overall max and min of trans
	int dynamicrange;
	int i;
	int compnum;
	int bilevel = 0;
	char coltype = 'M';	// monochrome
	// Make ymax equal to next highest multiple of blocksize
	int ymax = ((truerows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	// Make xmax equal to next highest multiple of 2*blocksize
	// Do this so that points can be processed in pairs in binary tree
	int xmax = ((truecols+(1<<(numlevels+1))-1)>>(numlevels+1))<<(numlevels+1);
	if ((xmax == padcols)&&(ymax == padrows)){ // Already padded.
		components[0] = incomponents[0];   // Compress in place
		components[1] = incomponents[1];
		components[2] = incomponents[2];
		}
	else
		pad(incomponents, components, truerows, truecols, pnmtype);
	
	gettrueminmax(components, pnmtype, minval, 
		maxval,truerows,truecols);
	if (quality == 100)
		from_bilevel(components, pnmtype, minval, maxval,
			ymax, xmax, bilevel);
	int spacing[65];	// Quantizer spacings
	int transform[9];
	if ((pnmtype == 3)&&(quality < 100))
		{
		coltype = 'Y';
		// rgb_ycc(components,ymax,xmax,1);
		calc_colour_transform(components,transform,ymax,xmax);
		rgb_pca(components,transform,ymax,xmax);
		gettrueminmax(components, pnmtype, tminval, tmaxval,
			truerows,truecols);
		}
	else {
		for (int i = 0; i < pnmtype; i++) {
			tminval[i] = minval[i];
			tmaxval[i] = maxval[i];
			}
		if (pnmtype == 3)
			coltype = 'G';	// GRB
		}
	omaxval = tmaxval[0];
	ominval = tminval[0];
	if (pnmtype == 3) {
		if (tmaxval[1] > omaxval) omaxval = tmaxval[1];
		if (tminval[1] < ominval) ominval = tminval[1];
		if (tmaxval[2] > omaxval) omaxval = tmaxval[2];
		if (tminval[2] < ominval) ominval = tminval[2];
		}
	if (bilevel) {
		omaxval = 1;
		ominval = 0;
		}
	dynamicrange = omaxval - ominval;
	if (calcq(dynamicrange, quality, spacing+1, numlevels*2) < 0) {
		return -1;
		}
	if (bilevel) {
		if (pnmtype == 1)
			coltype = 'B';	// Bilevel
		else
			coltype = 'C';	// Colour graphics bilevel
 		}
/*
	FILE *cfp = fopen(filename,"wb");
	if (cfp == 0) {
		return -2;
		}
*/
	// Write header
	if (pnmtype == 1)	
		fprintf(cfp,"btpc 5.a1%c%c%c0\n%d\n%d\n", 
			numlevels+'0', numcoders+'0', coltype, 
			truerows, truecols);
	else
		fprintf(cfp,"btpc 5.a3%c%c%c\n%d\n%d\n", 
			numlevels+'0', numcoders+'0', coltype, 
			truerows, truecols);
	if (comment) 
		fprintf(cfp,"%s",comment);
	fputc(0,cfp);	// String terminator
	if (coltype == 'Y')
		fprintf(cfp,"%d %d %d %d %d %d %d %d %d\n", transform[0],
		  transform[1], transform[2], transform[3], transform[4],
		  transform[5], transform[6], transform[7], transform[8]);

	// Output all the quantizer spacing parameters
	for (i = 1; i <=numlevels*2; i++)
		fputc((unsigned char) spacing[i], cfp);
	// And the minval and maxval parameters
	for (i = 0; i < pnmtype; i++) {
		fputc((unsigned char) minval[i], cfp);
		fputc((unsigned char) maxval[i], cfp);
		}
	fputc((unsigned char) ominval, cfp);
	fputc((unsigned char) omaxval, cfp);
	// Set up the huffman coders
	cfileio fio(cfp);
	rlcomp codeout[numcoders];
	for (i = 0; i < numcoders; i++) {
		// codeout[i].init();
		codeout[i].encodelength(xmax*ymax);	// Try to improve on
					// memory usage later.
		codeout[i].attach(&fio);
		codeout[i].indexnumber(i); // For debugging
#ifdef DEBUG
		codeout[i].debugon(); // For debugging
#endif
		}
	// Encode each plane of the picture
	makescratch(xmax, ymax);
	for (compnum = 0; compnum < pnmtype; compnum++) {
	    btpc_enc_plane(components[compnum],xmax,ymax,numlevels,
		0, spacing, codeout, ominval, omaxval,
		compnum, truecols, truerows);
	    }
	bincode(components, ymax, xmax, codeout, numlevels, spacing, 
		pnmtype, omaxval);
	fio.flush();
	fclose(cfp);
	// And finally, clean up
	if (components[0] != incomponents[0]) {
		for (compnum = 0; compnum < pnmtype; compnum++) {
	    		unsigned char **in = components[compnum];
	    		for (i = 0; i < ymax; i++)
				delete [] in[i];
	    		delete [] in;
	    		}
		}
	deletescratch(ymax);
	return 1;
	}

