//
// Binary tree predictive coding
// Macros used by enclevel.cpp and declevel.cpp
//
// Copyright (C) John Robinson, 1994, 1995, 1997, 2002.
//
#if defined(SQUARE) || defined(DIAMOND_LEFT) || defined(DIAMOND_RIGHT)
#undef P
#undef Q
#undef R
#undef S
#undef RIDGETHRESH
#undef HOK
#undef VOK
#undef HTEST
#undef VTEST
#undef SHAPE_H_EST
#undef SHAPE_V_EST
#undef SPACING
#undef RANGE
#undef ENCODE_TABLE
#undef DECODE_TABLE
#undef OUT_OF_RANGE
#undef OUTPEL
#undef VARIATION
#endif

#define VARIATION 32

#ifdef ENCODER
#undef C
#define C(X) (X)
#else
#undef C
#ifdef COMP0
// Using concatenation operator
#define C(X) (X ## 0)
#endif
#ifdef COMP1
#define C(X) (X ## 1)
#endif
#ifdef COMP2
#define C(X) (X ## 2)
#endif
#endif

#ifdef SQUARE
#define P (C(p))
#define Q (C(q))
#define R (C(r))
#define S (C(s))
#define RIDGETHRESH (sq_ridgethresh)
#define HOK ((X!=XHINT)&&(yplus!=YMAX))
#define VOK ((Y!=YHINT)&&(xplus!=XMAX))
#define HTEST ((X!=XHINT)&&(yplus!=YMAX)&& ((C(p)==C(pyminus)[xminus-XINTERVAL])&&\
		    (C(s)==C(pyplus)[xminus-XINTERVAL])))
#define VTEST ((Y!=YHINT)&&(xplus!=XMAX)&&((C(p)==C(py3minus)[xminus])&&\
		    (C(q)==C(py3minus)[xplus])))
#define SHAPE_H_EST (C(py)[X-XINTERVAL])
#define SHAPE_V_EST (C(py2minus)[X])
#define SPACING (sq_spacing)
#define RANGE (sq_range)
#define ENCODE_TABLE (sq_encode_table)
#define DECODE_TABLE (sq_decode_table)
#ifdef ENCODER
#define OUTPEL (prepel)
#else
#define OUTPEL (C(v))
#endif
#define OUT_OF_RANGE (sq_out_of_range)
#endif

#ifdef DIAMOND_LEFT
#define P (C(p))
#define Q (C(t))
#define R (C(q))
#define S (C(v))
#define RIDGETHRESH (di_ridgethresh)
#define HOK ((xplus!=XMAX)&&(yminus))
#define VOK ((xminus) &&(yminus))
#define HTEST ((xplus!=XMAX)&&(yminus)&& ((C(p)==C(py3minus)[xplus])&&\
                            (C(v)==C(py2minus)[X+XINTERVAL])))
#define VTEST ((xminus) &&(yminus)&& ((C(p)==C(py2minus)[X-XINTERVAL])&&\
                            (C(q)==C(py3minus)[xminus])))
#define SHAPE_H_EST (C(py2minus)[xplus])
#define SHAPE_V_EST (C(py2minus)[xminus])
#define SPACING (di_spacing)
#define ENCODE_TABLE (di_encode_table)
#ifdef ENCODER
#define OUTPEL (prepel)
#else
#define OUTPEL (tempout)
#endif
#define OUT_OF_RANGE (di_out_of_range)
#define DECODE_TABLE (di_decode_table)
#define RANGE (di_range)
#endif

#ifdef DIAMOND_RIGHT
#define P (C(u))
#define Q (C(p))
#define R (C(v))
#define S (C(s))
#define RIDGETHRESH (di_ridgethresh)
#define HOK ((xplus!=XMAX)&&(yminus))
#define VOK ((xminus) &&(yminus))
#define HTEST ((xplus!=XMAX)&&(yminus)&& ((C(u)==C(t))&&(C(s)==C(q))))
#define VTEST ((xminus) &&(yminus)&& ((C(u)==C(pyminus)[xminus-XINTERVAL])&&\
                            (C(p)==C(py2minus)[X-XINTERVAL])))
#define SHAPE_H_EST (C(pyminus)[X])
#define SHAPE_V_EST (C(pyminus)[X-XINTERVAL])
#define SPACING (di_spacing)
#define RANGE (di_range)
#define ENCODE_TABLE (di_encode_table)
#define DECODE_TABLE (di_decode_table)
#ifdef ENCODER
#define OUTPEL (prepel)
#else
#define OUTPEL (tempout)
#endif
#define OUT_OF_RANGE (di_out_of_range)
#endif

//
// Here is the predictor. For each pel, it's a set of condition tests on its
// neighbors, to decide which combination of them to use as the prediction.
// The neighbors form a square around the current pel, labelled clockwise from
// the top left, P, Q, R, S. See order.cpp for the orderarray definition,
// which has comments that refer to the signs of differences around the 
// square of neighbours. For example -0++ means
// P > Q, Q = R, R < S, S < P.
//
#if defined(SQUARE) || defined(DIAMOND_LEFT) || defined(DIAMOND_RIGHT)
#if defined(ENCODER) || defined(COMP0)
	if (comp == 0) {
		register int temp;
		register int order = 0;
		register int biggest, smallest;
		if ((temp=(Q-P)) > 0) {
			order = 2*27;
			biggest = Q;
			smallest = P;
			}
		else {
			if (temp == 0) 
				order = 27;
			biggest = P;
			smallest = Q;
			}
		if ((temp=(R-Q)) > 0) order += 2*9;
		else if (temp == 0) order += 9;
		if ((temp=(S-R)) > 0) order += 2*3;
		else if (temp == 0) order += 3;
		if ((temp=(P-S)) > 0) order += 2;
		else if (temp == 0) order += 1;
		predictor = orderarray[order];
		if (R > biggest) biggest = R;
		else if (R < smallest) smallest = R;
		if (S > biggest) biggest = S;
		else if (S < smallest) smallest = S;
		if ((biggest - smallest < RIDGETHRESH)&&
			((predictor < 3)||(predictor > 5)))
			// Second test is so we don't do this for completely
			// flat areas or graphic horiz and vert lines
			predictor = 0;
		}
#endif
	// printf("check x\n"); fflush(stdout);
	if (!predictor)
		estimate = (P+Q+R+S+2)>>2;
	else if (predictor < 5) {
		if (predictor == 1)
			estimate = (Q+S)>>1;
		else if (predictor == 2)
			estimate = (P+R)>>1;
		else if (predictor == 3)
			estimate = (VOK) ? SHAPE_V_EST : (P+Q)>>1;
		else // if (predictor == 4)
			estimate = (HOK) ? SHAPE_H_EST: (P+S)>>1;
		}
	else if (predictor < 11) {
		if (predictor == 5) {
			estimate = (VTEST) ? SHAPE_V_EST :
					(HTEST ? SHAPE_H_EST : P);
			}
		else if (predictor == 6) {
#ifdef ENCODER
			elow = (P+R)>>1; ehigh = (Q+S)>>1;
			// elow, ehigh not meaningful names any more
			if (abt[prepel-elow] < abt[ehigh-prepel]) {
				estimate = elow;
				ridgepel = 2;
				predictor = 2;
				}
			else {
				estimate = ehigh;
				ridgepel = 3;
				predictor = 1;
				}
#else
			if (codein[0].getbit()) {
				estimate = (Q+S)>>1;
				predictor = 1;
				}
			else {
				estimate = (P+R)>>1;
				predictor = 2;
				}
#endif
			}
		else if (predictor == 7)
			estimate = (Q+R)>>1;
		else if (predictor == 8)
			estimate = (S+R)>>1;
		else if (predictor == 9)
			estimate = (P+S)>>1;
		else // if (predictor == 10)
			estimate = (Q+P)>>1;
		}
	else {
		if (predictor == 11)
			estimate = P;
		else if (predictor == 12)
			estimate = R;
		else if (predictor == 13)
			estimate = Q;
		else // if (predictor == 14)
			estimate = S;
/*
		else {
			fprintf(stderr,"Error: predictor is %d\n", predictor);
			fprintf(stderr,"order is %d\n",order);
			exit(1);
			}
*/
		}
//
// In the encoder, go straight on to find the prediction error
//
	// printf("check y\n"); fflush(stdout);
#ifdef ENCODER
	// Now have a bit of inefficiency: going to check whether we're beyond
	// the true limits of the picture here and, if so, force the pel
	// value to equal the estimate, so the error will be 0.
#ifdef SQUARE
	if ((Y >= truerows)||(X >= truecols))
#elif DIAMOND_LEFT
	if ((yminus >= truerows)||(X >= truecols))
#else
	if ((Y >= truerows)||(xminus >= truecols))
#endif
		prepel = estimate;
	
        // At this point have estimate
	error = prepel - estimate;
        // Now quantize the error
        code = error = ENCODE_TABLE[error];
        if (code < 0)
                code += RANGE;
        // May cause wrap around for SPACING = 1

        // Finally change the value of the input pel to reflect what will
        // be recovered.
        OUTPEL = DECODE_TABLE[error] + estimate;
        if (OUTPEL < minval) OUTPEL = minval;
        else if (OUTPEL > maxval) OUTPEL = maxval;
#else
//
// In the decoder, propogate the trees then recover the value
//
#ifdef SQUARE
#elif DIAMOND_LEFT
	if (level > 1)
		pyqminus0[X-XQINT] = cpropagate;
#else	// DIAMOND_RIGHT
	if (level > 1)
		pyqplus0[X-XQINT] = dpropagate;
#endif
	    if (!error) {
		OUTPEL = estimate;
		}
            else
		{
               	OUTPEL =((error>OUT_OF_RANGE[estimate]) ?
	      	    myclip[DECODE_TABLE[error-RANGE]+estimate] :
               	    myclip[DECODE_TABLE[error]+estimate]);
		}
#ifdef SQUARE
	    C(py)[X] = OUTPEL;
#elif DIAMOND_LEFT
	    C(pyminus)[X] = OUTPEL;
#else	// DIAMOND_RIGHT
	    C(py)[xminus] = OUTPEL;
#endif

#endif
#ifdef DEBUG
	// printf("pred = %d, est = %d, error = %d, outpel = %d\n",predictor,estimate,error,OUTPEL);
#endif

#if defined(ENCODER) || defined(COMP0)
	// This is the first channel, so may change predictor
	// Inefficient tests here. Need to optimize this.
	error = abt[OUTPEL - estimate];	// In both encoder and decoder,
					// get the error again.
	if (!comp && error) {	
		register int temp;
		const int *offabt = &abt[-OUTPEL];
/*
		if ((temp=abt[OUTPEL - ((Q+S)>>1)]) < error) {
			error = temp;
			predictor = 1;
			}
		if ((temp=abt[OUTPEL - ((P+R)>>1)]) < error) {
			error = temp;
			predictor = 2;
			}
		if ((temp=abt[OUTPEL - ((Q+R)>>1)]) < error) {
			error = temp;
			predictor = 7;
			}
		if ((temp=abt[OUTPEL - ((S+R)>>1)]) < error) {
			error = temp;
			predictor = 8;
			}
		if ((temp=abt[OUTPEL - ((P+S)>>1)]) < error) {
			error = temp;
			predictor = 9;
			}
		if ((temp=abt[OUTPEL - ((P+Q)>>1)]) < error) {
			error = temp;
			predictor = 10;
			}
*/
	// printf("check z\n"); fflush(stdout);
		if ((temp=offabt[P]) < error) {
			error = temp;
			predictor = 11;
			}
		if ((temp=offabt[R]) < error) {
			error = temp;
			predictor = 12;
			}
		if ((temp=offabt[Q]) < error) {
			error = temp;
			predictor = 13;
			}
		if ((temp=offabt[S]) < error) {
			error = temp;
			predictor = 14;
			}
		if (VOK && ((temp=offabt[SHAPE_V_EST]) < error)) {
			error = temp;
			predictor = 3;
			}
		if (HOK && ((temp=offabt[SHAPE_H_EST]) < error)) {
			error = temp;
			predictor = 4;
			}
		}
#endif

#endif
